/***
*
* cmac.cc
*
* $Revision: 1.2 $
*
* Description:
*   Implementation of the CMAC class and supporting functions.
*
* Author:
*   Juan Carlos Santamaria  (carlos@cc.gatech.edu)
*     E-mail: carlos@cc.gatech.edu
*     URL:    http://www.cc.gatech.edu/ai/students/jcs
*
* File name:
*   $Id: cmac.cc,v 1.2 1996/09/19 22:17:48 carlos Exp $
*
* Revision History:
*   05/24/96    created.
*
*   $Log: cmac.cc,v $
*   Revision 1.2  1996/09/19  22:17:48  carlos
*   Fix replace_eligibilities and accumulate_eligibilities. The factor is
*   now 1/ngrids instead of 1. Fix compute_value_function as well.
*   This is only a conceptual fix since the final computation remains the
*   same.
*
*   Revision 1.1  1996/08/13  18:23:38  carlos
*   Initial revision
*
*
****/

#pragma  implementation


// -- Include files


#include <math.h>
#include <string.h>

#include "cmac.h"
#include "rand.h"


// -- Member function definitions

//============================================================================
// Variable::Variable

Variable::Variable( char nm[20], double min, double max, int ints )
{
    memset(name,0,sizeof(name));
    strncpy(name,nm,sizeof(name));

    max_limit = max;
    min_limit = min;
    range     = max - min;
    intervals = ints;
}


//============================================================================
// Wrap::get_tile

int Wrap::get_tile( double value, double offset )
{
    double tmp=fmod(value - offset - min_limit, range );
    tmp =  ( tmp>=0.0 ) ? min_limit + tmp  :  max_limit + tmp;
    return (int)((tmp-min_limit)/resolution);

}


//============================================================================
// Wrap::save

ostream& Wrap::save( ostream& to ) const
{
    to<<"name: "<<'\"'<<name<<'\"'<<endl;
    to<<"type: \"wrap\""<<endl;
    to<<"min: "<<min_limit<<endl;
    to<<"max: "<<max_limit<<endl;
    to<<"intervals: "<<intervals<<endl;

    return to;
}


//============================================================================
// Clip::get_tile

int Clip::get_tile( double value, double offset )
{
    double tmp = value<max_limit ? (value>min_limit ? value : min_limit) :
                                   max_limit;
    return (int)((tmp - min_limit - offset)/resolution);
}


//============================================================================
// Clip::save

ostream& Clip::save( ostream& to ) const
{
    to<<"name: "<<'\"'<<name<<'\"'<<endl;
    to<<"type: \"clip\""<<endl;
    to<<"max: "<<max_limit<<endl;
    to<<"min: "<<min_limit<<endl;
    to<<"intervals: "<<intervals<<endl;

    return to;
}


//============================================================================
// Grid::Grid

Grid::Grid( const CMAC& cmac, int nv, int v_ids[], double offs[]) : nvars(nv)
{
    int i;

    if ( nv>cmac.max_vars )
        error("Grid::Grid: too many variables");
    
    var_ids = new int [nvars];
    if ( var_ids==0 )
        error("Grid::Grid: cannot allocate variable array");
        
    offsets = new double [nvars];
    if ( offsets==0 )
        error("Grid::Grid: cannot allocate offsets array");
    
    for( i=0 ; i<nvars ; i++ ) {
        if ( (v_ids[i]>=0) && (v_ids[i]<cmac.nvars) )
            var_ids[i] = v_ids[i];
        else
            error("Grid::Grid: wrong variable id");
        offsets[i] = offs[i];
    }
}


//============================================================================
// Grid::~Grid

Grid::~Grid( void )
{
    delete [] var_ids;
    delete [] offsets; 
}


//============================================================================
// Grid::get_feature

int Grid::get_tile( const CMAC& cmac, double state[] )
{
    int       index;
    Variable *pvar;

    index = 0;
    for( int i=0 ; i<nvars ; i++ ) {
        pvar = cmac.vars[var_ids[i]];
        index *= pvar->intervals;
        index += pvar->get_tile( state[var_ids[i]], offsets[i] );
    }

    return index;
}


//============================================================================
// Grid::total_tiles

long Grid::total_tiles( const CMAC& cmac )
{
    long result = cmac.vars[var_ids[0]]->intervals;
    for( int i=1 ; i<nvars ; i++ )
        result *= cmac.vars[var_ids[i]]->intervals;
    return result;
}


//============================================================================
// Grid::save

ostream& Grid::save( ostream& to ) const
{
    to<<"\tvariables: "<<nvars<<endl;
    for(int i=0 ; i<nvars ; i++ ) {
        to<<"\tid: ";
        to.width(5);
        to<<var_ids[i];
        to.precision(25);
        to<<"\toffset: "<<offsets[i]<<endl;
    }

    return to;
}


//============================================================================
// CMAC::CMAC

CMAC::CMAC( int mx_v, int mx_g ) : max_vars(mx_v), max_grids(mx_g)
{
    vars = new Variable* [max_vars];
    if ( vars==0 )
        error("CMAC::CMAC: cannot allocate variable array");
    
    nvars=0;
    for( int i=0 ; i<max_vars ; i++ )
        vars[i]=0;
    
    grids = new Grid* [max_grids];
    if ( grids==0 )
        error("CMAC::CMAC: cannot allocate grids array");
    
    ngrids=0;
    for( int i=0 ; i<max_grids ; i++ )
        grids[i]=0;
}


//============================================================================
// CMAC::~CMAC

CMAC::~CMAC( void )
{
    delete [] vars;
    delete [] grids;
}


//============================================================================
// CMAC::define_variable

int CMAC::define_variable( char name[20], double min, double max,
                           int intervals, Type type )
{
    if ( nvars==max_vars )
        error("CMAC::define_variable: maximum number of variables reached");

    int id = nvars++;

    switch( type ) {
    case WRAP:
        vars[id] = new Wrap( name, min, max, intervals );
        break;
        
    case CLIP:
        vars[id] = new Clip( name, min, max, intervals );
        break;
    }
    
    return id;
}


//============================================================================
// CMAC::define_grid

int CMAC::define_grid( int nv, int ids[], double offs[] )
{
    if ( ngrids==max_grids )
        error("CMAC::define_grid: maximum number of grids reached");

    int id = ngrids++;
    
    grids[id] = new Grid( *this, nv, ids, offs );
    if ( grids[id]==0 ) 
        error("CMAC::define_grid: cannot allocate grid");
    
    return id;
}


//============================================================================
// CMAC::define_k_grids

void CMAC::define_k_grids( int k, int nv, int ids[] )
{
    double    offs[max_vars];
    Variable *pvar;
    int       i, j;
    
    for( i=0 ; i<k ; i++ ) {
        for( j=0 ; j<nv ; j++ ) {
            pvar = vars[ids[j]];
            offs[j] = (pvar->get_nominal_offset()-pvar->resolution*i)/k;
        }

        define_grid( nv, ids, offs );
    }
}


//============================================================================
// CMAC::total_tiles

long CMAC::total_tiles( void )
{
    long result = 0;
    for( int i=0 ; i<ngrids ; i++ )
        result += grids[0]->total_tiles(*this);
    return result;
}


//============================================================================
// CMAC::get_tiles

void CMAC::get_tiles( int tiles[], double state[] )
{
    for( int i=0 ; i<ngrids ; i++ ) 
        tiles[i] = grids[i]->get_tile( *this, state );
}


//============================================================================
// CMAC::save

ostream& CMAC::save( ostream& to ) const
{
    int i;
    
    // save variables

    to<<"variables: "<<nvars<<endl;
    for( i=0 ; i<nvars ; i++ )
        if ( !vars[i]->save( to ) )
            return to;

    // save grids

    to<<"grids: "<<ngrids<<endl;
    for( i=0 ; i<ngrids ; i++ )
        if ( !grids[i]->save( to ) )
            return to;

    return to;
}


//============================================================================
// CMAC::load

istream& CMAC::load( istream& from )
{
    char buffer[40];
    int  i, nv, ng;
    
    // load variables

    get_token( from, buffer, sizeof(buffer) );
    if ( !(from>>nv) )
        error("CMAC::load: error reading number of variables");
        
    for( i=0 ; i<nv ; i++ ) {
        char   name[20];
        Type   type;
        double max, min;
        int    intervals;
        int    fields = 0x1F;
        
        while( fields!=0 ) {
            get_token( from, buffer, sizeof(buffer) );
            if ( strcmp( buffer, "name" )== 0 ) {
                if ( !get_token(from,name,sizeof(name)) )
                    error("CMAC::load: error reading name",name);
                fields &= ~0x01;
            }
            else if ( strcmp( buffer, "type" )== 0 ) {
                if ( !get_token(from,buffer,sizeof(buffer)) )
                    error("CMAC::load: error reading type");

                if ( strcmp( buffer, "clip" ) == 0 )
                     type = CLIP;
                else if ( strcmp( buffer, "wrap" ) == 0 )
                     type = WRAP;
                else
                    error("CMAC::load: invalid file format in type: ",buffer);
                
                fields &= ~0x02;
            }
            else if ( strcmp( buffer, "min" )== 0 ) {
                if ( !(from>>min) )
                    error("CMAC::load: error reading min_limit");
                fields &= ~0x04;
            }
            else if ( strcmp( buffer, "max" ) == 0 ) {
                if ( !(from>>max) )
                    error("CMAC::load: error reading max_limit");
                fields &= ~0x08;
            }
            else if ( strcmp( buffer, "intervals" ) == 0 ) {
                if ( !(from>>intervals) )
                    error("CMAC::load: error reading intervals");
                fields &= ~0x10;
            }
            else {
                error("CMAC::load: invalid file format: ",buffer);
            }
        }

        define_variable(name,min,max,intervals,type);
    }

    // load grids

    get_token( from, buffer, sizeof(buffer) );
    if ( !(from>>ng) )
        error("CMAC::load: error reading number of grids");
        
    for( i=0 ; i<ng ; i++ ) {
        int    j;
        int    nv, ids[max_vars];
        double offs[max_vars];

        get_token( from, buffer, sizeof(buffer) );
        if ( !(from>>nv) )
            error("CMAC::load: error reading variables for grid");

        for( j=0 ; j<nv ; j++ ) {
            int    fields = 0x02;
        
            while( fields!=0 ) {
                get_token( from, buffer, sizeof(buffer) );
                if ( strcmp( buffer, "id" )== 0 ) {
                    if ( !(from>>ids[j]) )
                        error("CMAC::load: error reading id");
                    fields &= ~0x01;
                }
                else if ( strcmp( buffer, "offset" )== 0 ) {
                    if ( !(from>>offs[j]) )
                    error("CMAC::load: error reading offset");
                    fields &= ~0x02;
                }
                else {
                    error("CMAC::load: invalid file format: ",buffer);
                }
            }
        }

        define_grid(nv,ids,offs);
    }

    return from;
}


//============================================================================
// CMAC_CA::CMAC_CA

CMAC_CA::CMAC_CA( int mx_v, int mx_g ) : CMAC(mx_v,mx_g)
{
    cells = new Cell* [max_grids];
    if ( cells==0 )
        error("CMAC_CA::CMAC_CA: cannot allocate cells array");
    
    for( int i=0 ; i<max_grids ; i++ )
        cells[i]=0;
}


//============================================================================
// CMAC_CA::~CMAC_CA

CMAC_CA::~CMAC_CA( void )
{
    delete [] cells;
}


//============================================================================
// CMAC_CA::define_grid

int CMAC_CA::define_grid( int nv, int ids[], double offs[] )
{
    int id = CMAC::define_grid( nv, ids, offs);
    
    cells[id] = new Cell[ grids[id]->total_tiles(*this) ];
    if ( cells[id]==0 ) 
        error("CMAC_CA::define_grid: cannot allocate cells");

    return id;
}


//============================================================================
// CMAC_CA::initialize_value_function

void CMAC_CA::initialize_value_function( double value )
{
    int  ntiles, i, j;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        
        for( j=0 ; j<ntiles ; j++ )
            cells[i][j].weight = value/ngrids;
    }
}


//============================================================================
// CMAC_CA::compute_value_function

double CMAC_CA::compute_value_function( int tiles[] )
{
    double value=0.0;
    for( int i=0 ; i<ngrids ; i++ ) 
        value += cells[i][tiles[i]].weight;
    return value;
}


//============================================================================
// CMAC_CA::update_value_function

void CMAC_CA::update_value_function( double delta )
{
    int  i, j;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            cells[i][j].weight += cells[i][j].eligibility*delta;
    }
}
    

//============================================================================
// CMAC_CA::replace_eligibilities

void CMAC_CA::replace_eligibilities( int tiles[] )
{
    for( int i=0 ; i<ngrids ; i++ )
        cells[i][tiles[i]].eligibility = 1.0/ngrids;
}


//============================================================================
// CMAC_CA::accumulate_eligibilities

void CMAC_CA::accumulate_eligibilities( int tiles[] )
{
    for( int i=0 ; i<ngrids ; i++ )
        cells[i][tiles[i]].eligibility += 1.0/ngrids;
}


//============================================================================
// CMAC_CA::clear_eligibilities

void CMAC_CA::clear_eligibilities( void )
{
    int  i, j;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            cells[i][j].eligibility = 0.0;
    }
}


//============================================================================
// CMAC_CA::save

ostream& CMAC_CA::save( ostream& to ) const
{
    if ( !CMAC::save(to) )
        error("CMAC_CA::save: error writing CMAC_CA portion");

    int i, j, ntiles;
    for(i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        to<<"tilling: "<<ntiles<<endl;
        for(j=0 ; j<ntiles ; j++ ) {
            to.precision(25);
            to<<cells[i][j].weight<<' ';
        }
        if ( !(to<<endl) )
            error("CMAC_CA:save: error writing tilling weights");
    }

    return to;
}


//============================================================================
// CMAC_CA::load

istream& CMAC_CA::load( istream& from )
{
    if ( !CMAC::load(from) )
        error("CMAC_CA:load: error loading CMAC portion");

    char buffer[40];
    int  i, j, ntiles;
    for(i=0 ; i<ngrids ; i++ ) {
        get_token(from,buffer,sizeof(buffer));
        if ( !(from>>ntiles) )
            error("CMAC_CA:load: error reading number of tiles");
        
        for(j=0 ; j<ntiles ; j++ )
            if ( !(from>>cells[i][j].weight) )
                error("CMAC_CA:load: error reading weight");
    }

    return from;
}


//============================================================================
// CMAC_CA::update_eligibilities

void CMAC_CA::update_eligibilities( double gamma, double lambda )
{
    int  i, j;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            cells[i][j].eligibility *= gamma*lambda;
    }
}


//============================================================================
// CMAC_IA::CMAC_IA

CMAC_IA::CMAC_IA( int mx_v, int mx_g, int nm_a ) : CMAC(mx_v,mx_g), num_acts(nm_a)
{
    cells = new Cell** [max_grids];
    if ( cells==0 )
        error("CMAC_IA::CMAC_IA: cannot allocate cells array");
    
    for( int i=0 ; i<max_grids ; i++ )
        cells[i]=0;
}


//============================================================================
// CMAC_IA::~CMAC_IA

CMAC_IA::~CMAC_IA( void )
{
    delete [] cells;
}


//============================================================================
// CMAC_IA::define_grid

int CMAC_IA::define_grid( int nv, int ids[], double offs[] )
{
    int id     = CMAC::define_grid( nv, ids, offs);
    int ntiles = grids[id]->total_tiles(*this);
    
    cells[id] = new Cell* [ntiles];
    if ( cells[id]==0 ) 
        error("CMAC_IA::define_grid: cannot allocate cells");

    for( int j=0 ; j<ntiles ; j++ ) {
        cells[id][j] = new Cell[num_acts];
        if ( cells[id][j]==0 ) 
            error("CMAC_IA::define_grid: cannot allocate cells");
    }

    return id;
}


//============================================================================
// CMAC_IA::initialize_value_function

void CMAC_IA::initialize_value_function( double value )
{
    int  ntiles, i, j, k;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        
        for( j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ )
                cells[i][j][k].weight = value/ngrids;
    }
}


//============================================================================
// CMAC_IA::compute_value_function

double CMAC_IA::compute_value_function( int tiles[], int action_id )
{
    double value=0.0;
    for( int i=0 ; i<ngrids ; i++ ) 
        value += cells[i][tiles[i]][action_id].weight;
    return value;
}


//============================================================================
// CMAC_IA::update_value_function

void CMAC_IA::update_value_function( double delta )
{
    int  i, j, k;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ )
                cells[i][j][k].weight+=cells[i][j][k].eligibility*delta/ngrids;
    }
}
    

//============================================================================
// CMAC_IA::replace_eligibilities

void CMAC_IA::replace_eligibilities( int tiles[], int action_id )
{
    for( int i=0 ; i<ngrids ; i++ )
        for( int k=0 ; k<num_acts ; k++ )
            cells[i][tiles[i]][k].eligibility = (k==action_id) ? 1.0 : 0.0;
}


//============================================================================
// CMAC_IA::accumulate_eligibilities

void CMAC_IA::accumulate_eligibilities( int tiles[], int action_id )
{
    for( int i=0 ; i<ngrids ; i++ )
        cells[i][tiles[i]][action_id].eligibility += 1.0;
}


//============================================================================
// CMAC_IA::clear_eligibilities

void CMAC_IA::clear_eligibilities( void )
{
    int  i, j, k;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ )
                cells[i][j][k].eligibility = 0.0;
    }
}


//============================================================================
// CMAC_IA::save

ostream& CMAC_IA::save( ostream& to ) const
{
    if ( !CMAC::save(to) )
        error("CMAC_IA::save: error writing CMAC_IA portion");

    int i, j, k, ntiles;
    for(i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        to<<"tilling: "<<ntiles<<endl;
        for(j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ ) {
                to.precision(25);
                to<<cells[i][j][k].weight<<' ';
            }
        if ( !(to<<endl) )
            error("CMAC_IA:save: error writing tilling weights");
    }

    return to;
}


//============================================================================
// CMAC_IA::load

istream& CMAC_IA::load( istream& from )
{
    if ( !CMAC::load(from) )
        error("CMAC_IA:load: error loading CMAC portion");

    char buffer[40];
    int  i, j, k, ntiles;
    for(i=0 ; i<ngrids ; i++ ) {
        get_token(from,buffer,sizeof(buffer));
        if ( !(from>>ntiles) )
            error("CMAC_IA:load: error reading number of tiles");
        
        for(j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ ) 
                if ( !(from>>cells[i][j][k].weight) )
                    error("CMAC_IA:load: error reading weight");
    }

    return from;
}


//============================================================================
// CMAC_IA::update_eligibilities

void CMAC_IA::update_eligibilities( double gamma, double lambda )
{
    int  i, j, k;
    long ntiles;

    for( i=0 ; i<ngrids ; i++ ) {
        ntiles = grids[i]->total_tiles(*this);
        for( j=0 ; j<ntiles ; j++ )
            for( k=0 ; k<num_acts ; k++ )
                cells[i][j][k].eligibility *= gamma*lambda;
    }
}


//============================================================================
// Table::Table

Table::Table( int mx_v ) : CMAC(mx_v,1)
{
    weights=0;
}


//============================================================================
// Table::~Table

Table::~Table( void )
{
    delete [] weights;
}


//============================================================================
// Table::define_grid

int Table::define_grid( int nv, int ids[], double offs[] )
{
    int id = CMAC::define_grid( nv, ids, offs);
    
    weights = new double[ grids[id]->total_tiles(*this) ];
    if ( weights==0 ) 
        error("Table::define_grid: cannot allocate weights");

    return id;
}


// -- Global function definitions

//============================================================================
// select_best

int select_best( int nvals, double values[], double percent, int strm )
{
    double max;
    int    best[nvals];
    int    n;

    if ( percent==1.0 || simrand(95)<percent ) {
        max=values[0];
        for( int j=1 ; j<nvals ; j++ )
            if ( values[j]>max )
                max=values[j];

        n=0;
        for( int j=0 ; j<nvals ; j++ )
            if ( values[j]==max )
                best[n++]=j;

        if ( n>1 ) 
            return best[(int)uniform(0.0,n,strm)];

        return best[0];
    }
    else
        return (int)uniform(0,nvals,strm);
}


//============================================================================
// save_cmac

bool save_cmac( char filename[30], const CMAC& cmac )
{
    ofstream to(filename,ios::out );
    
    if (!to) return FALSE;
    
    // save header

    to<<"max_vars: "<<cmac.max_vars<<endl;
    to<<"max_grids: "<<cmac.max_grids<<endl;
    
    return cmac.save(to) ? TRUE : FALSE;
}


//============================================================================
// load_cmac

CMAC* load_cmac( char filename[30] )
{
    ifstream from(filename,ios::in | ios::nocreate );
    char     buffer[40];
    int      fields = 0x02;
    int      max_vars, max_grids;
    
    if (!from) return (CMAC*)0;

    while ( fields!=0 ) {
        if ( !get_token(from, buffer, sizeof(buffer) ))
           error("load_cmac: error in file");
        
        if ( strcmp( buffer, "max_vars")== 0 ) {
            if ( !(from>>max_vars) )
                error("load_cmac: error reading max_vars");
            fields &= ~0x01;
        }
        else if ( strcmp( buffer, "max_grids")== 0 ) {
            if ( !(from>>max_grids) )
                error("load_cmac: error reading max_grids");
            fields &= ~0x02;
        }
    }

    CMAC *pcmac = new CMAC(max_vars,max_grids);
    if ( pcmac==0 )
        error("load_cmac: cannot allocate cmac");

    return pcmac->load( from ) ? pcmac : (delete pcmac, (CMAC *)0);
}



//============================================================================
// save_cmac_ca

bool save_cmac_ca( char filename[30], const CMAC_CA& cmac )
{
    ofstream to(filename,ios::out );
    
    // save header

    to<<"max_vars: "<<cmac.max_vars<<endl;
    to<<"max_grids: "<<cmac.max_grids<<endl;
    
    return cmac.save(to) ? TRUE : FALSE;
}


//============================================================================
// load_cmac_ca

CMAC_CA* load_cmac_ca( char filename[30] )
{
    ifstream from(filename,ios::in | ios::nocreate );
    char     buffer[40];
    int      fields = 0x02;
    int      max_vars, max_grids;

    while ( fields!=0 ) {
        if ( !get_token(from, buffer, sizeof(buffer) ))
           error("load_cmac_ca: error in file");
        
        if ( strcmp( buffer, "max_vars")== 0 ) {
            if ( !(from>>max_vars) )
                error("load_cmac_ca: error reading max_vars");
            fields &= ~0x01;
        }
        else if ( strcmp( buffer, "max_grids")== 0 ) {
            if ( !(from>>max_grids) )
                error("load_cmac_ca: error reading max_grids");
            fields &= ~0x02;
        }
    }

    CMAC_CA *pcmac = new CMAC_CA(max_vars,max_grids);
    if ( pcmac==0 )
        error("load_cmac_ca: error allocating cmac");

    return pcmac->load( from ) ? pcmac : (delete pcmac, (CMAC_CA *)0);
}


//============================================================================
// save_cmac_ia

bool save_cmac_ia( char filename[30], const CMAC_IA& cmac )
{
    ofstream to(filename,ios::out );
    
    // save header

    to<<"max_vars: "<<cmac.max_vars<<endl;
    to<<"max_grids: "<<cmac.max_grids<<endl;
    to<<"num_acts: "<<cmac.num_acts<<endl;
    
    return cmac.save(to) ? TRUE : FALSE;
}


//============================================================================
// load_cmac_ia

CMAC_IA* load_cmac_ia( char filename[30] )
{
    ifstream from(filename,ios::in | ios::nocreate );
    char     buffer[40];
    int      fields = 0x07;
    int      max_vars, max_grids, num_acts;

    while ( fields!=0 ) {
        if ( !get_token(from, buffer, sizeof(buffer) ))
           error("load_cmac_ia: error in file");
        
        if ( strcmp( buffer, "max_vars")== 0 ) {
            if ( !(from>>max_vars) )
                error("load_cmac_ia: error reading max_vars");
            fields &= ~0x01;
        }
        else if ( strcmp( buffer, "max_grids")== 0 ) {
            if ( !(from>>max_grids) )
                error("load_cmac_ia: error reading max_grids");
            fields &= ~0x02;
        }
        else if ( strcmp( buffer, "num_acts")== 0 ) {
            if ( !(from>>num_acts) )
                error("load_cmac_ia: error reading num_acts");
            fields &= ~0x04;
        }
    }

    CMAC_IA *pcmac = new CMAC_IA(max_vars,max_grids,num_acts);
    if ( pcmac==0 )
        error("load_cmac_ia: error allocating cmac");

    return pcmac->load( from ) ? pcmac : (delete pcmac, (CMAC_IA *)0);
}


/****************************** end of file *********************************/

