/***
*
* cmac.h
*
* $Revision: 1.1 $
*
* Description:
*   Interface for the CMAC class and supporting functions.
*   The class CMAC is an abstract class that supports the basic functions
*   common to all CMACs.
*   The class CMAC_CA implements SARSA using continuous actions.
*   The class CMAC-IA implements SARSA using indexed actions.
*
* Author:
*   Juan Carlos Santamaria
*     E-mail: carlos@cc.gatech.edu
*     URL:    http://www.cc.gatech.edu/ai/students/jcs
*
* File name:
*   $Id: cmac.h,v 1.1 1996/08/13 18:23:05 carlos Exp $
*
* Revision History:
*   $Log: cmac.h,v $
*   Revision 1.1  1996/08/13  18:23:05  carlos
*   Initial revision
*
*
*****************************************************************************/


#ifndef _INCL_CMAC
#define _INCL_CMAC


// -- Include files

#include <iostream.h>
#include <fstream.h>

#ifndef _INCL_USRDEF
#include "usrdef.h"
#endif


// -- Class declarations

class CMAC;
class CMAC_CA;
class CMAC_IA;
class Table;


class Variable {
public:
    char   name[20];
    double max_limit;
    double min_limit;
    int    intervals;
    double range;
    double resolution;

    Variable( char name[20], double max, double min, int intvs );
    
    virtual ~Variable( void ) {}

    virtual int    get_tile( double value, double offset ) = 0;
    virtual double get_nominal_offset( void ) = 0;

    virtual ostream& save( ostream& to ) const = 0;
};


class Wrap : public Variable {
    friend class CMAC;
    
    Wrap( char name[20], double max, double min, int intvs ) :
        Variable(name,max,min,intvs) { resolution=range/intvs; }
        
    int    get_tile( double value, double offset );
    double get_nominal_offset( void ) { return 0.0; }

    ostream& save( ostream& to ) const;
};


class Clip : public Variable {
    friend class CMAC;
    
    Clip( char name[20], double max, double min, int intvs ) :
        Variable(name,max,min,intvs) { resolution=range/(intvs-1); }
        
    int    get_tile( double value, double offset );
    double get_nominal_offset( void ) { return -resolution/2.0; }

    ostream& save( ostream& to ) const;
};


class Grid {
    friend class CMAC;
    friend class CMAC_CA;
    friend class CMAC_IA;
    friend class Table;
    
    const int nvars;
    
    int     *var_ids;
    double  *offsets;

    Grid( const CMAC& cmac, int nvars, int var_ids[], double offsets[] );
    ~Grid( void );

    int  get_tile( const CMAC& cmac, double state[] );

    long total_tiles( const CMAC& cmac );

    ostream& save( ostream& to ) const;
};


class Cell {
    friend class CMAC_CA;
    friend class CMAC_IA;

    double weight;
    double eligibility;

    Cell( double w=0.0, double e=0.0 ) : weight(w), eligibility(e) {}
};


class CMAC {
public:
    enum Type { CLIP, WRAP };

    const int  max_vars;
    const int  max_grids;

    int        nvars;
    Variable **vars;

    int        ngrids;
    Grid     **grids;

    CMAC( int mx_v = 1, int mx_g = 1 );
    virtual ~CMAC( void );
    
    virtual int  define_variable( char name[20], double min, double max,
                                  int intervals, Type type );

    virtual int  define_grid( int nvars, int var_ids[], double offsets[] );
    virtual void define_k_grids( int k, int nvars, int var_ids[] );

    long total_tiles( void );
    
    void get_tiles( int tiles[], double state[] );

    virtual ostream& save( ostream& to ) const;
    virtual istream& load( istream& from );

    friend bool  save_cmac( char filename[30], const CMAC& cmac );
    friend CMAC* load_cmac( char filename[30] );
};


class CMAC_CA : public CMAC {
public:
    Cell **cells;

    CMAC_CA( int max_vars, int max_grids );
    ~CMAC_CA( void );

    int    define_grid( int nvars, int var_ids[], double offsets[] );
    
    void   initialize_value_function( double value );
    double compute_value_function( int tiles[] );
    void   update_value_function( double delta );
    
    void   replace_eligibilities( int tiles[] );
    void   accumulate_eligibilities( int tiles[] );
    void   clear_eligibilities( void );
    void   update_eligibilities( double gamma, double lambda );

    ostream& save( ostream& to ) const;
    istream& load( istream& from );

    friend bool     save_cmac_ca( char filename[30], const CMAC_CA& cmac );
    friend CMAC_CA* load_cmac_ca( char filename[30] );
};


class CMAC_IA : public CMAC {
public:
    const int num_acts;
    
    Cell ***cells;

    CMAC_IA( int max_vars, int max_grids, int num_actions );
    ~CMAC_IA( void );

    int    define_grid( int nvars, int var_ids[], double offsets[] );
    
    void   initialize_value_function( double value );
    double compute_value_function( int tiles[], int action_id );
    void   update_value_function( double delta );
    
    void   replace_eligibilities( int tiles[], int action_id );
    void   accumulate_eligibilities( int tiles[], int action_id );
    void   clear_eligibilities( void );
    void   update_eligibilities( double gamma, double lambda );

    ostream& save( ostream& to ) const;
    istream& load( istream& from );

    friend bool     save_cmac_ia( char filename[30], const CMAC_IA& cmac );
    friend CMAC_IA* load_cmac_ia( char filename[30] );
};


class Table : public CMAC {
public:
    double *weights;

    Table( int max_vars );
    ~Table( void );

    int define_grid( int nvars, int var_ids[], double offsets[] );

    ostream& save( ostream& to ) const { return to; }
    istream& load( istream& from )     { return from; }

    friend bool   save_table( char filename[30], const Table& table );
    friend Table* load_table( char filename[30] );
};


// -- Function prototypes

int  select_best( int nvals, double values[], double percent, int strm );


#endif
